package app.crossword.yourealwaysbe.puz;

import app.crossword.yourealwaysbe.puz.Playboard.Position;

import java.io.Serializable;
import java.time.LocalDate;
import java.util.Arrays;
import java.util.List;
import java.util.LinkedList;
import java.util.Objects;

import java.util.logging.Logger;

public class Puzzle implements Serializable{
    private static final Logger LOG = Logger.getLogger("app.crossword.yourealwaysbe");

    private String author;
    private String copyright;
    private String notes;
    private String title;
    private String[] acrossClues;
    private Integer[] acrossCluesLookup;
    private String[] downClues;
    private Integer[] downCluesLookup;
    private int numberOfClues;
    private LocalDate pubdate = LocalDate.now();
    private String source;
    private String sourceUrl = "";
    private Box[][] boxes;
    private Box[] boxesList;
    private String[] rawClues;
    private boolean updatable;
    private int height;
    private int width;
    private long playedTime;
    private boolean scrambled;
    public short solutionChecksum;
    private String version;
    private boolean hasGEXT;
    private Position position;
    private boolean across = true;

    private Note[] acrossNotes = null;
    private Note[] downNotes = null;

    private LinkedList<HistoryItem> historyList = new LinkedList<>();

    // Temporary fields used for unscrambling.
    public int[] unscrambleKey;
    public byte[] unscrambleTmp;
    public byte[] unscrambleBuf;

    public void setAcrossClues(String[] acrossClues) {
        this.acrossClues = acrossClues;
    }

    public String[] getAcrossClues() {
        return acrossClues;
    }

    public void setAcrossCluesLookup(Integer[] acrossCluesLookup) {
        this.acrossCluesLookup = acrossCluesLookup;
    }

    public Integer[] getAcrossCluesLookup() {
        return acrossCluesLookup;
    }

    public void setAuthor(String author) {
        this.author = author;
    }

    public String getAuthor() {
        return author;
    }

    private static Box checkedGet(Box[][] boxes, int row, int col) {
        try {
            return boxes[row][col];
        } catch (IndexOutOfBoundsException e) {
            return null;
        }
    }

    private static boolean somethingAbove(Box[][] boxes, int row, int col) {
        return checkedGet(boxes, row - 1, col) != null;
    }

    private static boolean somethingBelow(Box[][]boxes, int row, int col) {
        return checkedGet(boxes, row + 1, col) != null;
    }

    private static boolean somethingLeftOf(Box[][] boxes, int row, int col) {
        return checkedGet(boxes, row, col - 1) != null;
    }

    private static boolean somethingRightOf(Box[][] boxes, int row, int col) {
        return checkedGet(boxes, row, col + 1) != null;
    }

    public void setBoxes(Box[][] boxes) {
        this.boxes = boxes;

        int clueCount = 1;

        for (int row = 0; row < boxes.length; row++) {
            for (int col = 0; col < boxes[row].length; col++) {
                if (boxes[row][col] == null) {
                    continue;
                }

                boolean tickedClue = false;
                if (!somethingAbove(boxes, row, col)) {
                    if (somethingBelow(boxes, row, col)) {
                        boxes[row][col].setDown(true);
                        tickedClue = true;
                    }
                }

                if (!somethingLeftOf(boxes, row, col)) {
                    if (somethingRightOf(boxes, row, col)) {
                        boxes[row][col].setAcross(true);
                        tickedClue = true;
                    }
                }

                if (tickedClue) {
                    boxes[row][col].setClueNumber(clueCount);
                    clueCount++;
                }
            }
        }

        // set partOfAcross numbers
        int maxRowLen = -1;
        for (int row = 0; row < boxes.length; row++) {
            int lastAcross = -1;
            int acrossPosition = -1;
            maxRowLen = java.lang.Math.max(maxRowLen, boxes[row].length);
            for (int col = 0; col < boxes[row].length; col++) {
                if (boxes[row][col] == null) {
                    lastAcross = -1;
                    continue;
                }

                if (boxes[row][col].isAcross()) {
                    lastAcross = boxes[row][col].getClueNumber();
                    acrossPosition = 0;
                }

                if (lastAcross > 0) {
                    boxes[row][col].setPartOfAcrossClueNumber(lastAcross);
                    boxes[row][col].setAcrossPosition(acrossPosition);
                    acrossPosition++;
                }
            }
        }

        // set partOfDown numbers
        for (int col = 0; col < maxRowLen; col++) {
            int lastDown = -1;
            int downPosition = -1;
            for (int row = 0; row < boxes.length; row++) {
                if (col >= boxes[row].length || boxes[row][col] == null) {
                    lastDown = -1;
                    continue;
                }

                if (boxes[row][col].isDown()) {
                    lastDown = boxes[row][col].getClueNumber();
                    downPosition = 0;
                }

                if (lastDown > -1) {
                    boxes[row][col].setPartOfDownClueNumber(lastDown);
                    boxes[row][col].setDownPosition(downPosition);
                    downPosition++;
                }
            }
        }

    }

    public Box[][] getBoxes() {
        return (boxes == null) ? this.buildBoxes() : boxes;
    }

    public void setBoxesList(Box[] value) {
        this.boxesList = value;
    }

    public Box[] getBoxesList() {
        Box[] result = new Box[boxes.length * boxes[0].length];
        int i = 0;

        for (int x = 0; x < boxes.length; x++) {
            for (int y = 0; y < boxes[x].length; y++) {
                result[i++] = boxes[x][y];
            }
        }

        return result;
    }

    /**
     * Initialize the temporary unscramble buffers.  Returns the scrambled solution.
     */
    public byte[] initializeUnscrambleData() {
    	unscrambleKey = new int[4];
    	unscrambleTmp = new byte[9];

    	byte[] solution = getSolutionDown();
    	unscrambleBuf = new byte[solution.length];

    	return solution;
    }



	private byte[] getSolutionDown() {
		StringBuilder ans = new StringBuilder();
		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				if (boxes[y][x] != null) {
					ans.append(boxes[y][x].getSolution());
				}
			}
		}
		return ans.toString().getBytes();
	}

	public void setUnscrambledSolution(byte[] solution) {
		int i = 0;
		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				if (boxes[y][x] != null) {
					boxes[y][x].setSolution((char) solution[i++]);
				}
			}
		}
		setScrambled(false);
		setUpdatable(false);
	}

    public void setCopyright(String copyright) {
        this.copyright = copyright;
    }

    public String getCopyright() {
        return copyright;
    }

    public void setDate(LocalDate date) {
        this.pubdate = date;
    }

    public LocalDate getDate() {
        return pubdate;
    }

    public void setDownClues(String[] downClues) {
        this.downClues = downClues;
    }

    public String[] getDownClues() {
        return downClues;
    }

    public void setDownCluesLookup(Integer[] downCluesLookup) {
        this.downCluesLookup = downCluesLookup;
    }

    public Integer[] getDownCluesLookup() {
        return downCluesLookup;
    }

    /**
     * @param height the height to set
     */
    public void setHeight(int height) {
        this.height = height;
    }

    /**
     * @return the height
     */
    public int getHeight() {
        return height;
    }

    public void setNotes(String notes) {
        this.notes = notes;
    }

    public String getNotes() {
        return notes;
    }

    public void setNumberOfClues(int numberOfClues) {
        this.numberOfClues = numberOfClues;
    }

    public int getNumberOfClues() {
        return numberOfClues;
    }

    public int getPercentComplete() {
        int total = 0;
        int correct = 0;

        for (int x = 0; x < boxes.length; x++) {
            for (int y = 0; y < boxes[x].length; y++) {
                if (boxes[x][y] != null) {
                    total++;

                    if (boxes[x][y].getResponse() == boxes[x][y].getSolution()) {
                        correct++;
                    }
                }
            }
        }
        if(total == 0){
            return 0;
        }
        return (correct * 100) / (total);
    }

    public int getPercentFilled() {
        int total = 0;
        int filled = 0;

        for (int x = 0; x < boxes.length; x++) {
            for (int y = 0; y < boxes[x].length; y++) {
                if (boxes[x][y] != null) {
                    total++;

                    if (!boxes[x][y].isBlank()) {
                        filled++;
                    }
                }
            }
        }

        return (filled * 100) / (total);
    }

    public void setRawClues(String[] rawClues) {
        this.rawClues = rawClues;
    }

    public String[] getRawClues() {
        return rawClues;
    }

    public void setSource(String source) {
        this.source = source;
    }

    public String getSource() {
        return source;
    }

    public void setSourceUrl(String sourceUrl) {
        this.sourceUrl = sourceUrl;
    }

    public String getSourceUrl() {
        return sourceUrl;
    }

    public void setTime(long time) {
        this.playedTime = time;
    }

    public long getTime() {
        return this.playedTime;
    }

    public void setTitle(String title) {
        this.title = title;
    }

    public String getTitle() {
        return title;
    }

    public void setUpdatable(boolean updatable) {
        this.updatable = updatable;
    }

    public boolean isUpdatable() {
        return updatable;
    }

    public void setVersion(String version) {
    	this.version = version;
    }

     public String getVersion() {
    	 return version;
     }

    public void setGEXT(boolean hasGEXT) {
    	this.hasGEXT = hasGEXT;
    }

    public boolean getGEXT() {
    	return hasGEXT;
    }

    public void setPosition(Position position) {
    	this.position = position;
    }

    public Position getPosition() {
    	return position;
    }

    public void setAcross(boolean across) {
    	this.across = across;
    }

    public boolean getAcross() {
    	return across;
    }

    public void setScrambled(boolean scrambled) {
    	this.scrambled = scrambled;
    }

    public boolean isScrambled() {
    	return scrambled;
    }

    public void setSolutionChecksum(short checksum) {
    	this.solutionChecksum = checksum;
    }

    public short getSolutionChecksum() {
    	return solutionChecksum;
    }

    /**
     * @param width the width to set
     */
    public void setWidth(int width) {
        this.width = width;
    }

    /**
     * @return the width
     */
    public int getWidth() {
        return width;
    }

    public Box[][] buildBoxes() {
        int i = 0;
        boxes = new Box[this.height][this.width];

        for (int y = 0; y < this.height; y++) {
            for (int x = 0; x < this.width; x++) {
                boxes[y][x] = boxesList[i++];
            }
        }

        return boxes;
    }

    public Note[] getAcrossNotes() {
        return acrossNotes;
    }

    public Note[] getDownNotes() {
        return downNotes;
    }

    public Note getNote(int clueNum, boolean isAcross) {
        if (isAcross && acrossNotes == null)
            return null;
        else if (!isAcross && downNotes == null)
            return null;

        if (isAcross) {
            int idx = getAcrossClueIndex(clueNum);
            return (idx < 0) ? null : acrossNotes[idx];
        } else {
            int idx = getDownClueIndex(clueNum);
            return (idx < 0) ? null : downNotes[idx];
        }
    }

    /**
     * Assumes acrossClues and downClues has been initialised
     */
    public void setNote(Note note, int clueNum, boolean isAcross) {
        int idx = isAcross ? getAcrossClueIndex(clueNum)
                           : getDownClueIndex(clueNum);

        setNoteRaw(note, idx, isAcross);
    }

    /**
     * Assumes acrossClues and downClues has been initialised
     */
    public void setNoteRaw(Note note, int clueIdx, boolean isAcross) {
        if (isAcross) {
            if (clueIdx >= 0 && clueIdx <= acrossClues.length) {
                if (acrossNotes == null) {
                    acrossNotes = new Note[acrossClues.length];
                }
                acrossNotes[clueIdx] = note;
            }
        } else {
            if (clueIdx >= 0 && clueIdx <= downClues.length) {
                if (downNotes == null) {
                    downNotes = new Note[downClues.length];
                }
                downNotes[clueIdx] = note;
            }
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }

        if (obj == null) {
            return false;
        }

        if (Puzzle.class != obj.getClass()) {
            return false;
        }

        Puzzle other = (Puzzle) obj;

        if (!Arrays.equals(acrossClues, other.acrossClues)) {
            //            for(int i=0; i < acrossClues.length; i++)
            //            	System.out.println((acrossClues[i].equals(other.acrossClues[i]))+"["+acrossClues[i]+"]==["+other.acrossClues[i]+"]");
            return false;
        }

        if (!Arrays.equals(acrossCluesLookup, other.acrossCluesLookup)) {
            return false;
        }

        if (author == null) {
            if (other.author != null) {
                return false;
            }
        } else if (!author.equals(other.author)) {
            return false;
        }

        Box[][] b1 = boxes;
        Box[][] b2 = other.boxes;
        boolean boxEq = true;

        for (int x = 0; x < b1.length; x++) {
            for (int y = 0; y < b1[x].length; y++) {
                boxEq = boxEq
                    ? ((b1[x][y] == b2[x][y]) || b1[x][y].equals(b2[x][y]))
                    : boxEq;
            }
        }

        if (!boxEq) {
            return false;
        }

        if (copyright == null) {
            if (other.copyright != null) {
                return false;
            }
        } else if (!copyright.equals(other.copyright)) {
            return false;
        }

        if (!Arrays.equals(downClues, other.downClues)) {
            return false;
        }

        if (!Arrays.equals(downCluesLookup, other.downCluesLookup)) {
            return false;
        }

        if (height != other.height) {
            return false;
        }

        if (notes == null) {
            if (other.notes != null) {
                return false;
            }
        } else if (!notes.equals(other.notes)) {
            return false;
        }

        if (getNumberOfClues() != other.getNumberOfClues()) {
            return false;
        }

        if (title == null) {
            if (other.title != null) {
                return false;
            }
        } else if (!title.equals(other.title)) {
            return false;
        }

        if (width != other.width) {
            return false;
        }

        if (version == null) {
	        if (other.version != null) {
	            return false;
	        }
        } else if (!version.equals(other.version)) {
            return false;
        }

        if (scrambled != other.scrambled) {
        	return false;
        }

        if (solutionChecksum != other.solutionChecksum) {
        	return false;
        }

        if (!Arrays.equals(acrossNotes, other.acrossNotes)) {
            return false;
        }

        if (!Arrays.equals(downNotes, other.downNotes)) {
            return false;
        }

        return true;
    }

    public String findAcrossClue(int clueNumber) {
        return this.acrossClues[getAcrossClueIndex(clueNumber)];
    }

    public String findDownClue(int clueNumber) {
        return this.downClues[getDownClueIndex(clueNumber)];
    }

    public int getAcrossClueIndex(int number) {
        return Arrays.binarySearch(this.acrossCluesLookup, number);
    }

    public int getDownClueIndex(int number) {
        return Arrays.binarySearch(this.downCluesLookup, number);
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = (prime * result) + Arrays.hashCode(acrossClues);
        result = (prime * result) + Arrays.hashCode(acrossCluesLookup);
        result = (prime * result) + ((author == null) ? 0 : author.hashCode());
        result = (prime * result) + Arrays.hashCode(boxes);
        result = (prime * result) +
            ((copyright == null) ? 0 : copyright.hashCode());
        result = (prime * result) + Arrays.hashCode(downClues);
        result = (prime * result) + Arrays.hashCode(downCluesLookup);
        result = (prime * result) + height;
        result = (prime * result) + ((notes == null) ? 0 : notes.hashCode());
        result = (prime * result) + getNumberOfClues();
        result = (prime * result) + ((title == null) ? 0 : title.hashCode());
        result = (prime * result) + ((version == null) ? 0 : version.hashCode());
        result = (prime * result) + width;
        result = (prime *result) + Arrays.hashCode(acrossNotes);
        result = (prime *result) + Arrays.hashCode(downNotes);

        return result;
    }

    @Override
    public String toString() {
        return "Puzzle " + boxes.length + " x " + boxes[0].length + " " +
        this.title;
    }

    public void updateHistory(int clueNumber, boolean across) {
        int clueIndex = across ?
                        getAcrossClueIndex(clueNumber) :
                        getDownClueIndex(clueNumber);

        if (clueIndex > -1) {
            HistoryItem item = new HistoryItem(clueNumber, across);
            // if a new item, not equal to most recent
            if (historyList.isEmpty() ||
                !item.equals(historyList.getFirst())) {
                historyList.remove(item);
                historyList.addFirst(item);
            }
        }
    }

    public void setHistory(List<HistoryItem> newHistory) {
        historyList.clear();
        for (HistoryItem item : newHistory) {
            int number = item.getClueNumber();
            int clueIndex = item.getAcross() ?
                            getAcrossClueIndex(number) :
                            getDownClueIndex(number);
            if (clueIndex > -1)
                historyList.add(item);
        }
    }

    public List<HistoryItem> getHistory() {
        return historyList;
    }

    public static class HistoryItem {
        private int clueNumber;
        private boolean across;

        public HistoryItem(int clueNumber, boolean across) {
            this.clueNumber = clueNumber;
            this.across = across;
        }

        public int getClueNumber() { return clueNumber; }
        public boolean getAcross() { return across; }

        public boolean equals(Object o) {
            if (o instanceof HistoryItem) {
                HistoryItem other = (HistoryItem) o;
                return clueNumber == other.clueNumber && across == other.across;
            }
            return false;
        }

        public int hashCode() {
            return Objects.hash(clueNumber, across);
        }

        public String toString() {
            return clueNumber + (across ? "a" : "d");
        }
    }
}
